import logging
import pytz

from datetime import datetime
from psi_elt import get_logger
from psi_elt.abstract_elt_job import AbstractELTJob
from typeguard import typechecked
from typing import Any, Dict


@typechecked
class AbstractRunner:
    def __init__(
        self,
        jobs: Dict[str, Any],
        logger: logging.Logger = None
    ):
        '''
        Initializes an AbstractRunner.

        This class should be treated as abstract and should be extended.
        Child classes must override the _run function. Optionally, child
        classes may override the parse_event and validate_parsed_event
        functions as well.

        This class includes utilties for getting a capture time and
        initializing an ELT job.
        '''
        assert jobs, 'A jobs dict is required to initialize an AbstractRunner'
        self.__jobs = jobs
        self.__logger = logger or get_logger()

    @property
    def logger(self) -> logging.Logger:
        return self.__logger

    def run(self, event: Dict[str, Any]) -> Dict[str, Any]:
        t0 = datetime.now()
        parsed_event = self.parse_event(event)
        self.validate_parsed_event(parsed_event)
        self.logger.info(f'Running with parsed event: {parsed_event}')
        job = self.get_job(
            job_name=parsed_event.pop('name'),
            capture_time=parsed_event['capture_time'],
            job_init_kwargs=parsed_event.pop('job_init_kwargs')
        )
        is_test = parsed_event.pop('is_test')
        s3_uri = self._run(job, is_test, **parsed_event)
        self.logger.info(f'ELT Job completed in {str(datetime.now() - t0)}')
        return {
            'event': event,
            'parsed_event': parsed_event,
            'job': job.to_dict(),
            'results': {
                's3_uri': s3_uri
            }
        }

    def _run(self, job: AbstractELTJob, **parsed_event):
        raise Exception('AbstractRunner._run should be overridden by the child class')

    def parse_event(self, event: Dict[str, Any]) -> Dict[str, Any]:
        raise Exception('AbstractRunner.parse_event should be overridden by the child class')

    def validate_parsed_event(self, parsed_event: Dict[str, Any]):
        raise Exception('AbstractRunner.validate_parsed_event should be overridden by the child class')

    def get_capture_time(self, event: dict) -> datetime:
        if 'capture_time' in event and event['capture_time']:
            # fromisoformat returns a naive datetime object that IS NOT adjusted for the local timezone
            return datetime.fromisoformat(event['capture_time']).replace(tzinfo=pytz.utc)
        elif isinstance(event.get('s3_source'), dict) and event['s3_source'].get('s3_uri'):
            capture_time_ms = event['s3_source']['s3_uri'].split('/')[-1]
            # fromtimestamp returns a naive datetime object that IS adjusted for the local timezone
            return datetime.fromtimestamp(int(capture_time_ms)/1000).astimezone(pytz.utc)
        elif 's3_uri' in event and event['s3_uri']:
            capture_time_ms = event['s3_uri'].split('/')[-1]
            # fromtimestamp returns a naive datetime object that IS adjusted for the local timezone
            return datetime.fromtimestamp(int(capture_time_ms)/1000).astimezone(pytz.utc)
        else:
            # now returns a naive datetime object that IS adjusted for the local timezone
            return datetime.now().astimezone(pytz.utc)

    def get_job(
        self,
        job_name: str,
        capture_time: datetime,
        job_init_kwargs: dict
    ) -> AbstractELTJob:
        assert capture_time, 'A capture_time is required to initialize an ELT job'
        cls = self.__jobs.get(job_name)
        if cls:
            return cls(
                capture_time,
                **job_init_kwargs
            )
        raise Exception(f'Unrecognized ELT job with name: {job_name}')



import boto3
import json
import pytz
import re
import logging

from datetime import datetime, time
from psi_elt import get_logger
from typeguard import typechecked
from typing import Any, Dict

EXTRACTOR_ELT_JOB_TYPE = 'extractor'
LOADER_ELT_JOB_TYPE = 'loader'
TRANSFORMER_ELT_JOB_TYPE = 'transformer'
ELT_JOB_TYPES = [EXTRACTOR_ELT_JOB_TYPE, LOADER_ELT_JOB_TYPE, TRANSFORMER_ELT_JOB_TYPE]


@typechecked
class AbstractELTJob:
    def __init__(
        self,
        name: str,
        elt_job_type: str,
        capture_time: datetime = None,
        aws_session: Any = None,
        logger: logging.Logger = None
    ):
        assert name, 'A name is required to initialize an AbstractELTJob'
        assert elt_job_type in ELT_JOB_TYPES, f'Invalid ELT job type: {elt_job_type}'
        self.__name = name
        self.__elt_job_type = elt_job_type
        self.__aws_session = aws_session or boto3.Session()
        self.__logger = logger or get_logger()

        # Default capture time of 6am UTC
        if capture_time is None:
            self.logger.warning('No capture time provided, using now as the capture time')
            self.__capture_time = datetime.now().astimezone(pytz.utc)
        elif not capture_time.tzinfo:
            self.__capture_time = capture_time.replace(tzinfo=pytz.utc)
        else:
            self.__capture_time = capture_time.astimezone(pytz.utc)
        self.__default_capture_time = datetime.combine(
            self.__capture_time.date(),
            time(hour=6, tzinfo=pytz.utc)
        )
        self.__default_capture_time_ms = self.dt_to_ms(self.__default_capture_time)
        self.logger.info(f'Initialized ELT job: {self}')

    @property
    def name(self) -> str:
        return self.__name

    @property
    def elt_job_type(self) -> str:
        return self.__elt_job_type

    @property
    def capture_time(self) -> datetime:
        return self.__capture_time

    @property
    def default_capture_time(self) -> datetime:
        return self.__default_capture_time

    @property
    def default_capture_time_ms(self) -> int:
        return self.__default_capture_time_ms

    @property
    def aws_session(self) -> Any:
        return self.__aws_session

    @property
    def logger(self) -> logging.Logger:
        return self.__logger

    def get_s3_object(self, s3_uri: str) -> Any:
        assert s3_uri

        s3_path = s3_uri.split('s3://')[-1]
        bucket = s3_path.split('/', 1)[0]
        key = s3_path.split('/', 1)[-1]

        try:
            client = self.aws_session.client('s3')
            response = client.get_object(Bucket=bucket, Key=key)
            return response['Body'].read().decode('utf-8')
        except Exception as e:
            self.logger.error(f'Failed to get s3 object {s3_path} with exception: {e}')
            raise e

    def get_parameter(self, parameter: str) -> str:
        assert parameter
        try:
            client = self.aws_session.client('ssm')
            response = client.get_parameter(Name=parameter, WithDecryption=True)
            value = response.get('Parameter', {}).get('Value')
            if value is None:
                raise Exception('Parameter value not found in response')
            return value
        except Exception as e:
            self.logger.error(f'Failed to get parameter {parameter} with exception: {e}')
            raise e

    def format_date(self, dt: datetime) -> str:
        '''Formats the date as a string. If the provided datetime is naive (no timezone), this
        assumes UTC and returns a string like "2021-05-26T01:02:03Z". If the provided datetime has
        a non-UTC timezone, returns a string like "2021-05-26T01:02:03-05:00".
        '''
        if dt.tzinfo and dt.tzinfo != pytz.UTC:
            return dt.isoformat(timespec='seconds')
        return dt.strftime('%Y-%m-%dT%H:%M:%SZ')

    def parse_date(self, dt: str) -> datetime:
        '''Parses the given string into a datetime. By default, expects UTC offset information in
        the given string (e.g. "2021-05-26T01:02:03-05:00") and assigns the appropriate time zone.
        If dt is a string like "2021-05-26T01:02:03Z", assumes the timezone is UTC.
        '''
        try:
            if dt.endswith('Z'):
                parsed = datetime.strptime(dt, '%Y-%m-%dT%H:%M:%SZ')
                return pytz.timezone('UTC').localize(parsed)
            return datetime.strptime(''.join(dt.rsplit(':', 1)), '%Y-%m-%dT%H:%M:%S%z')
        except Exception as e:
            raise ValueError(
                'Datetime must be a string like "2021-05-26T01:02:03Z" for UTC times or "2021-05-26T01:02:03-05:00" for non-UTC times'
            )

    def is_tappid_compliant(self, t_appid: str) -> bool:
        return bool(re.match('^SVC[0-9]{5}$', t_appid))

    def dt_to_ms(self, dt: datetime) -> int:
        if dt:
            return int(round(dt.timestamp() * 1000))
        return -1

    def to_dict(self) -> Dict[str, Any]:
        return {
            'name': self.name,
            'elt_job_type': self.elt_job_type,
            'capture_time': str(self.capture_time),
            'default_capture_time': str(self.default_capture_time),
            'default_capture_time_ms': self.default_capture_time_ms
        }

    def __repr__(self) -> str:
        return json.dumps(self.to_dict())



import logging

from datetime import datetime
from psi_elt.abstract_runner import AbstractRunner
from psi_elt.custom_extractor import CustomExtractor
from typeguard import typechecked
from typing import Any, Dict


@typechecked
class ExtractorRunner(AbstractRunner):
    def __init__(
        self,
        extractors: Dict[str, Any],
        s3_loader: Any = None,
        logger: logging.Logger = None
    ):
        '''
        Initializes a ExtractorRunner.

        Accepts a extractors dict, mapping extractor name to extractor class.
        For example:
        {
            "divvy": DivvyExtractor,
            "pagerduty": PagerDutyExtractor
        }
        '''
        super().__init__(jobs=extractors, logger=logger)
        assert s3_loader, 'A S3 loader is required to initialize an ExtractorRunner'
        self.__s3_loader = s3_loader

    def parse_event(self, event: Dict[str, Any]) -> Dict[str, Any]:
        '''Parses an extractor event like:
        {
            'name'          : 'extractor-name',   # required
            'capture_time'  : '2022-01-01T06:00',
            'is_test'       : False,
            'write_to_s3'   : False,
            'account_id'    : '012345678901',
            'bucket_name'   : 'my-s3-bucket',
            'output_s3_path': 'my/s3/path'
        }
        Required event keys:
           1. name         : the name of the extractor to run

        Optional event keys:
           1. capture_time   : target time for the job to run against
           2. is_test        : boolean for whethor or not this is a test run
           3. write_to_s3    : boolean for whether or not to write to S3
           4. account_id     : AWS account ID to extract data from
           5. bucket_name    : S3 bucket to write data to
           6. output_s3_path : S3 path to write data to

        Returns a parsed event like:
        {
            'name'               : 'extractor-name',
            'capture_time'       : datetime(2022, 1, 1, hour=6, tzinfo=pytz.utc),
            'job_init_kwargs'    : {},
            'is_test'            : False,
            'write_to_s3'        : False,
            'extract_kwargs'     : {
                'account_id'     : '012345678901'
            },
            's3_load_kwargs'     : {
                'bucket_name'    : 'my-s3-bucket',
                'output_s3_path' : 'my/s3/path'
            }
        }
        Child classes may override this function to implement custom event parsing rules.
        '''
        parsed_event = {
            'name': event.get('name'),
            'capture_time': self.get_capture_time(event),
            'job_init_kwargs': {},
            'is_test': bool(event.get('is_test')),
            'write_to_s3': event.get('write_to_s3', False),
            'extract_kwargs': {},
            's3_load_kwargs': {}
        }
        if event.get('account_id'):
            parsed_event['extract_kwargs']['account_id'] = event.get('account_id')
        if event.get('bucket_name'):
            parsed_event['s3_load_kwargs']['bucket_name'] = event.get('bucket_name')
        if event.get('output_s3_path'):
            parsed_event['s3_load_kwargs']['output_s3_path'] = event.get('output_s3_path')
        return parsed_event

    def validate_parsed_event(self, parsed_event: Dict[str, Any]):
        '''Validates that a parsed event has the required format'''
        required_keys = [
            'name', 'capture_time', 'job_init_kwargs', 'is_test', 'write_to_s3', 'extract_kwargs',
            's3_load_kwargs'
        ]
        missing_keys = [k for k in required_keys if k not in parsed_event]
        extra_keys = [k for k in parsed_event if k not in required_keys]
        assert not missing_keys, f'Parsed event is missing keys: {missing_keys}'
        assert not extra_keys, f'Parsed event has extra keys: {extra_keys}'
        assert isinstance(parsed_event['name'], str), f"Invalid name, expected a string but got: {parsed_event['name']}"
        assert isinstance(parsed_event['capture_time'], datetime), f"Invalid capture_time, expected a datetime but got: {parsed_event['capture_time']}"
        assert isinstance(parsed_event['job_init_kwargs'], dict), f"Invalid job_init_kwargs, expected a dict but got: {parsed_event['job_init_kwargs']}"
        assert isinstance(parsed_event['is_test'], bool), f"Invalid is_test, expected a bool but got: {parsed_event['is_test']}"
        assert isinstance(parsed_event['write_to_s3'], bool), f"Invalid write_to_s3, expected a bool but got: {parsed_event['write_to_s3']}"
        assert isinstance(parsed_event['extract_kwargs'], dict), f"Invalid extract_kwargs, expected a dict but got: {parsed_event['extract_kwargs']}"
        assert isinstance(parsed_event['s3_load_kwargs'], dict), f"Invalid s3_load_kwargs, expected a dict but got: {parsed_event['s3_load_kwargs']}"
        
        extract_kwargs = parsed_event['extract_kwargs']
        if 'account_id' in extract_kwargs:
            assert type(extract_kwargs['account_id']) in [type(None), str], f"Invalid extract_kwargs.account_id, expected a string but got: {extract_kwargs['account_id']}"

        s3_load_kwargs = parsed_event['s3_load_kwargs']
        if 'bucket_name' in s3_load_kwargs:
            assert isinstance(s3_load_kwargs['bucket_name'], str), f"Invalid s3_load_kwargs.bucket_name, expected a string but got: {s3_load_kwargs['bucket_name']}"
        if 'output_s3_path' in s3_load_kwargs:
            assert isinstance(s3_load_kwargs['output_s3_path'], str), f"Invalid s3_load_kwargs.output_s3_path, expected a string but got: {s3_load_kwargs['output_s3_path']}"

    def _run(
        self,
        extractor: CustomExtractor,
        is_test: bool,
        capture_time: datetime,
        write_to_s3: bool,
        extract_kwargs: dict,
        s3_load_kwargs: dict
    ) -> Any:
        data = extractor.extract(is_test, extract_kwargs)
        s3_uri = None
        if write_to_s3:
            self.logger.info('Beginning upload to S3')
            s3_uri = self.__s3_loader.load(
                extractor.name,
                data,
                capture_time,
                **s3_load_kwargs
            )
        return s3_uri
